@@ -1,10 +1,10 @@
 /***** BEGIN LICENSE BLOCK *****
- * Version: CPL 1.0/GPL 2.0/LGPL 2.1
+ * Version: EPL 1.0/GPL 2.0/LGPL 2.1
  *
- * The contents of this file are subject to the Common Public
+ * The contents of this file are subject to the Eclipse Public
  * License Version 1.0 (the "License"); you may not use this file
  * except in compliance with the License. You may obtain a copy of
- * the License at http://www.eclipse.org/legal/cpl-v10.html
+ * the License at http://www.eclipse.org/legal/epl-v10.html
  *
  * Software distributed under the License is distributed on an "AS
  * IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
@@ -23,11 +23,11 @@
  * in which case the provisions of the GPL or the LGPL are applicable instead
  * of those above. If you wish to allow use of your version of this file only
  * under the terms of either the GPL or the LGPL, and not to allow others to
- * use your version of this file under the terms of the CPL, indicate your
+ * use your version of this file under the terms of the EPL, indicate your
  * decision by deleting the provisions above and replace them with the notice
  * and other provisions required by the GPL or the LGPL. If you do not delete
  * the provisions above, a recipient may use your version of this file under
- * the terms of any one of the CPL, the GPL or the LGPL.
+ * the terms of any one of the EPL, the GPL or the LGPL.
  ***** END LICENSE BLOCK *****/
 package org.jruby.util;
 
@@ -107,7 +107,7 @@ public final class ByteList implements Comparable, CharSequence, Serializable {
     public ByteList(byte[] bytes, Encoding encoding) {
         this.bytes = bytes;
         this.realSize = bytes.length;
-        this.encoding = encoding;
+        this.encoding = safeEncoding(encoding);
     }
 
     /**
@@ -129,13 +129,27 @@ public final class ByteList implements Comparable, CharSequence, Serializable {
      * @param copy whether to arraycopy wrap for the backing store or not
      */
     public ByteList(byte[] wrap, boolean copy) {
+        this(wrap, ASCIIEncoding.INSTANCE, copy);
+    }
+
+    /**
+     * Create a new instance of ByteList with the contents of wrap.  If copy is true then it will
+     * array copy the contents.  Otherwise it will use the byte array passed in as its initial
+     * backing store.
+     *
+     * @param wrap the initial bytes for this ByteList
+     * @param encoding the encoding for the bytes
+     * @param copy whether to arraycopy wrap for the backing store or not
+     */
+    public ByteList(byte[] wrap, Encoding encoding, boolean copy) {
         assert wrap != null;
         if (copy) {
-            bytes = (byte[])wrap.clone();
+            this.bytes = (byte[])wrap.clone();
         } else {
-            bytes = wrap;
+            this.bytes = wrap;
         }
-        realSize = wrap.length;
+        this.realSize = wrap.length;
+        this.encoding = safeEncoding(encoding);
     }
 
     /**
@@ -145,7 +159,7 @@ public final class ByteList implements Comparable, CharSequence, Serializable {
      * @param wrap is contents for new ByteList
      */
     public ByteList(ByteList wrap) {
-        this(wrap.bytes, wrap.begin, wrap.realSize);
+        this(wrap.bytes, wrap.begin, wrap.realSize, wrap.encoding, true);
     }
 
     /**
@@ -164,7 +178,7 @@ public final class ByteList implements Comparable, CharSequence, Serializable {
      */
     @Deprecated
     public ByteList(ByteList wrap, boolean copy) {
-        this(wrap.bytes, wrap.begin, wrap.realSize, false);
+        this(wrap.bytes, wrap.begin, wrap.realSize, wrap.encoding, false);
     }
 
     /**
@@ -192,17 +206,33 @@ public final class ByteList implements Comparable, CharSequence, Serializable {
      */
     // FIXME:  Fix the index != 0 not honoring copy and separate out into a different caller. JRuby.next would be the right time for this.
     public ByteList(byte[] wrap, int index, int len, boolean copy) {
+        this(wrap, index, len, ASCIIEncoding.INSTANCE, copy);
+    }
+
+    /**
+     * Create a new instance of ByteList using wrap as a backing store where index is the first
+     * index in the byte array where the data starts and len indicates how long the data portion
+     * of the bytelist is.  wrap will be array copied if copy is true OR if index != 0.
+     *
+     * @param wrap the bytes to use
+     * @param index where in the bytes the data starts
+     * @param len how long the data is in the wrap array
+     * @param copy if true array copy wrap. otherwise use as backing store
+     */
+    public ByteList(byte[] wrap, int index, int len, Encoding encoding, boolean copy) {
         assert wrap != null : "'wrap' must not be null";
         assert index >= 0 && index <= wrap.length : "'index' is not without bounds of 'wrap' array";
         assert wrap.length >= index + len : "'index' + 'len' is longer than the 'wrap' array";
 
-        if (copy || index != 0) {
+        if (copy) {
             bytes = new byte[len];
             System.arraycopy(wrap, index, bytes, 0, len);
         } else {
+            begin = index;
             bytes = wrap;
         }
         realSize = len;
+        this.encoding = safeEncoding(encoding);
     }
 
     /**
@@ -233,6 +263,7 @@ public final class ByteList implements Comparable, CharSequence, Serializable {
         realSize -= len;
         
         System.arraycopy(bytes, start + len, bytes, start, realSize);
+        invalidate();
     }
 
     /**
@@ -246,6 +277,7 @@ public final class ByteList implements Comparable, CharSequence, Serializable {
         for ( ; --len >= 0; ) {
             append(b);
         }
+        invalidate();
     }
 
     /**
@@ -276,7 +308,7 @@ public final class ByteList implements Comparable, CharSequence, Serializable {
         ByteList dup = new  ByteList(bytes, false);
         dup.realSize = realSize;
         dup.begin = begin;
-        dup.encoding = encoding;
+        dup.encoding = safeEncoding(encoding);
         dup.hash = hash;
         dup.stringValue = stringValue;
         return dup;
@@ -292,7 +324,7 @@ public final class ByteList implements Comparable, CharSequence, Serializable {
         ByteList dup = new ByteList(length);
 
         dup.append(this.bytes, this.begin, this.realSize);
-        dup.encoding = encoding;
+        dup.encoding = safeEncoding(encoding);
         
         return dup;
     }
@@ -304,10 +336,12 @@ public final class ByteList implements Comparable, CharSequence, Serializable {
      * @param length to use to make sure ByteList is long enough
      */
     public void ensure(int length) {
-        if (length >= bytes.length) {
-            byte[] tmp = new byte[length + (length >>> 1)];
+        if (begin + length > bytes.length) {
+            byte[] tmp = new byte[Math.min(Integer.MAX_VALUE, length + (length >>> 1))];
             System.arraycopy(bytes, begin, tmp, 0, realSize);
             bytes = tmp;
+            begin = 0;
+            invalidate();
         }
     }
 
@@ -340,6 +374,7 @@ public final class ByteList implements Comparable, CharSequence, Serializable {
     public void view(int index, int len) {
         realSize = len;
         begin = begin + index;
+        invalidate();
     }
 
     /**
@@ -382,9 +417,10 @@ public final class ByteList implements Comparable, CharSequence, Serializable {
      */
     public void prepend(byte b) {
         grow(1);
-        System.arraycopy(bytes, 0, bytes, 1, realSize);
-        bytes[0] = b;
+        System.arraycopy(bytes, begin + 0, bytes, begin + 1, realSize);
+        bytes[begin + 0] = b;
         realSize++;
+        invalidate();
     }
 
     /**
@@ -395,7 +431,9 @@ public final class ByteList implements Comparable, CharSequence, Serializable {
      */
     public ByteList append(byte b) {
         grow(1);
-        bytes[realSize++] = b;
+        bytes[begin + realSize] = b;
+        realSize++;
+        invalidate();
         return this;
     }
 
@@ -423,8 +461,9 @@ public final class ByteList implements Comparable, CharSequence, Serializable {
         grow(length);
         int read = 0;
         int n;
+        int start = begin + realSize;
         while (read < length) {
-            n = input.read(bytes, begin + read, length - read);
+            n = input.read(bytes, start + read, length - read);
             if (n == -1) {
                 if (read == 0) throw new java.io.EOFException();
                 break;
@@ -433,6 +472,7 @@ public final class ByteList implements Comparable, CharSequence, Serializable {
         }
 
         realSize += read;
+        invalidate();
         return this;
     }
 
@@ -445,8 +485,9 @@ public final class ByteList implements Comparable, CharSequence, Serializable {
      */
     public void append(ByteBuffer buffer, int len) {
         grow(len);
-        buffer.get(bytes, realSize, len);
+        buffer.get(bytes, begin + realSize, len);
         realSize += len;
+        invalidate();
     }
 
     /**
@@ -458,8 +499,9 @@ public final class ByteList implements Comparable, CharSequence, Serializable {
         assert moreBytes != null : "moreBytes is null";
 
         grow(moreBytes.length);
-        System.arraycopy(moreBytes, 0, bytes, realSize, moreBytes.length);
+        System.arraycopy(moreBytes, 0, bytes, begin + realSize, moreBytes.length);
         realSize += moreBytes.length;
+        invalidate();
     }
 
     /**
@@ -498,8 +540,9 @@ public final class ByteList implements Comparable, CharSequence, Serializable {
         assert len >= 0 && moreBytes.length - start >= len : "Bad length";
 
         grow(len);
-        System.arraycopy(moreBytes, start, bytes, realSize, len);
+        System.arraycopy(moreBytes, start, bytes, begin + realSize, len);
         realSize += len;
+        invalidate();
     }
 
     /**
@@ -516,6 +559,7 @@ public final class ByteList implements Comparable, CharSequence, Serializable {
         byte tmp[] = new byte[length];
         System.arraycopy(bytes, 0, tmp, 0, realSize);
         bytes = tmp;
+        invalidate();
     }
 
     /**
@@ -582,21 +626,7 @@ public final class ByteList implements Comparable, CharSequence, Serializable {
         assert begin + index < begin + realSize : "index is too large";
         
         bytes[begin + index] = (byte)b;
-    }
-
-    /**
-     * Replace the byte array backing store with newBytes.  This method is only referred to in
-     * deprecated method RubyString.view(CharSequence).
-     *
-     * We deprecated this method because it ignore begin and if we
-     *
-     * @param newBytes
-     */
-    @Deprecated
-    public void replace(byte[] newBytes) {
-        assert newBytes != null;
-        this.bytes = newBytes;
-        realSize = newBytes.length;
+        invalidate();
     }
 
     /**
@@ -625,39 +655,61 @@ public final class ByteList implements Comparable, CharSequence, Serializable {
         System.arraycopy(bytes,beg+len,bytes,beg+count,realSize - (len+beg));
         System.arraycopy(nbytes,index,bytes,beg,count);
         realSize = newSize;
+        invalidate();
     }
 
     /**
-     * Note: This is as unsafe as unsafeReplace
+     * Replace all bytes in this ByteList with bytes from the given array.
      *
-     * @param beg
-     * @param len
-     * @param nbytes
+     * @param source bytes to use for replacement
      */
-    public void replace(int beg, int len, ByteList nbytes) {
-        replace(beg, len, nbytes.bytes, nbytes.begin, nbytes.realSize);
+    public void replace(byte[] source) {
+        replace(0, realSize, source, 0, source.length);
     }
 
     /**
-     * Note: This is as unsafe as unsafeReplace
-     * @param beg
-     * @param len
-     * @param buf
+     * Replace a region of bytes in this ByteList with bytes from the given ByteList.
+     *
+     * @param targetOff offset of target region in this ByteList
+     * @param targetLen length of target region in this ByteList
+     * @param source ByteList to use for replacement
      */
-    public void replace(int beg, int len, byte[] buf) {
-        replace(beg, len, buf, 0, buf.length);
+    public void replace(int targetOff, int targetLen, ByteList source) {
+        replace(targetOff, targetLen, source.bytes, source.begin, source.realSize);
     }
 
     /**
-     * Note: This is as unsafe as unsafeReplace
-     * @param beg
-     * @param len
-     * @param nbytes
-     * @param index
-     * @param count
+     * Replace a region of bytes in this ByteList with bytes from the given array.
+     *
+     * @param targetOff offset of target region in this ByteList
+     * @param targetLen length of target region in this ByteList
+     * @param source bytes to use for replacement
      */
-    public void replace(int beg, int len, byte[] nbytes, int index, int count) {
-        unsafeReplace(beg, len, nbytes, index, count);
+    public void replace(int targetOff, int targetLen, byte[] source) {
+        replace(targetOff, targetLen, source, 0, source.length);
+    }
+
+    /**
+     * Replace a region of bytes in this ByteList with a region of bytes from the given array.
+     *
+     * @param targetOff offset of target region in this ByteList
+     * @param targetLen length of target region in this ByteList
+     * @param source bytes to use for replacement
+     * @param sourceOff offset of source region in the replacement bytes
+     * @param sourceLen length of source region in the replacement bytes
+     */
+    public void replace(int targetOff, int targetLen, byte[] source, int sourceOff, int sourceLen) {
+        int newSize = realSize - targetLen + sourceLen;
+        ensure(newSize);
+        int tailSize = realSize - (targetLen + targetOff);
+        if (tailSize != 0) {
+            System.arraycopy(bytes,begin+targetOff+targetLen,bytes,begin+targetOff+sourceLen,tailSize);
+        }
+        if (sourceLen != 0) {
+            System.arraycopy(source,sourceOff,bytes,begin+targetOff,sourceLen);
+        }
+        realSize = newSize;
+        invalidate();
     }
 
     public void insert(int index, int b) {
@@ -665,6 +717,7 @@ public final class ByteList implements Comparable, CharSequence, Serializable {
         System.arraycopy(bytes, index, bytes, index + 1, realSize - index);
         bytes[index] = (byte)b;
         realSize++;
+        invalidate();
     }
 
     /**
@@ -840,7 +893,7 @@ public final class ByteList implements Comparable, CharSequence, Serializable {
     }
 
     public boolean startsWith(ByteList other, int toffset) {
-        if (realSize == 0) return false;
+        if (realSize == 0 || this.realSize < other.realSize + toffset) return false;
 
         byte[]ta = bytes;
         int to = begin + toffset;
@@ -1039,11 +1092,15 @@ public final class ByteList implements Comparable, CharSequence, Serializable {
     private void grow(int increaseRequested) {
         if (increaseRequested < 0) return;
 
+        // new available size
         int newSize = realSize + increaseRequested;
-        if (newSize > bytes.length) {
+
+        // only recopy if bytes does not have enough room *after* the begin index
+        if (newSize > bytes.length - begin) {
             byte[] newBytes = new byte[newSize + (newSize >> 1)];
-            if (bytes.length != 0) System.arraycopy(bytes, 0, newBytes, 0, realSize);
+            if (bytes.length != 0) System.arraycopy(bytes, begin, newBytes, 0, realSize);
             bytes = newBytes;
+            begin = 0;
         }
     }
 
@@ -1052,9 +1109,13 @@ public final class ByteList implements Comparable, CharSequence, Serializable {
      */
     @Override
     public int hashCode() {
-        if (hash != 0) return hash;
+        int currentHash = hash;
+        if (currentHash != 0) return currentHash;
 
         int key = 0;
+        int begin = this.begin, realSize = this.realSize;
+        byte[] bytes = this.bytes;
+        
         int index = begin;
         final int end = begin + realSize;
         while (index < end) {
@@ -1072,10 +1133,11 @@ public final class ByteList implements Comparable, CharSequence, Serializable {
      */
     @Override
     public String toString() {
-        if (stringValue == null) {
-            stringValue = decode(bytes, begin, realSize, "ISO-8859-1");
+        String decoded = this.stringValue;
+        if (decoded == null) {
+            this.stringValue = decoded = decode(bytes, begin, realSize, "ISO-8859-1");
         }
-        return stringValue;
+        return decoded;
     }
 
     /**
@@ -1273,6 +1335,7 @@ public final class ByteList implements Comparable, CharSequence, Serializable {
     public final void setUnsafeBytes(byte[] bytes) {
         assert bytes != null;
         this.bytes = bytes;
+        invalidate();
     }
 
     /**
@@ -1288,6 +1351,7 @@ public final class ByteList implements Comparable, CharSequence, Serializable {
     public final void setBegin(int begin) {
         assert begin >= 0;
         this.begin = begin;
+        invalidate();
     }
 
     /**
@@ -1303,6 +1367,7 @@ public final class ByteList implements Comparable, CharSequence, Serializable {
     public final void setRealSize(int realSize) {
         assert realSize >= 0;
         this.realSize = realSize;
+        invalidate();
     }
 
     /**
@@ -1317,6 +1382,15 @@ public final class ByteList implements Comparable, CharSequence, Serializable {
      */
     public final void setEncoding(Encoding encoding) {
         assert encoding != null;
-        this.encoding = encoding;
+        this.encoding = safeEncoding(encoding);
+        invalidate();
+    }
+
+    /**
+     * Ensure the encoding is always non-null.
+     */
+    public static Encoding safeEncoding(Encoding incoming) {
+        if (incoming == null) return ASCIIEncoding.INSTANCE;
+        return incoming;
     }
 }
